from pygame import Vector2, image
from game import Game, GameObject
from universum import Universum
import pygame
import numpy as np
from utils import Timer


class ViewRect:
    def __init__(self, dim):
        self.dim = dim
        self.topleft = np.zeros(dim)
        self.botright = np.zeros(dim)

    def set_size(self, size):
        self.topleft += np.array(size)
    
    def move(self, offset):
        self.topleft += np.array(offset)
        self.botright += np.array(offset)
    
    def zoom(self, val):
        dif = (self.botright - self.topleft) / 2
        self.topleft -= dif * val
        self.botright += dif * val

    def size(self):
        return np.abs(self.botright - self.topleft)
    
    def diagonal(self):
        return np.sqrt(np.sum(self.size() ** 2))
    
    def __repr__(self) -> str:
        return f"<ViewRect start={self.topleft}, end={self.botright}>"


timer = Timer()


class UniversObject(GameObject):
    def __init__(self, universum: Universum) -> None:
        self.universum = universum
        self.view_rect = ViewRect(2)
        self.view_rect.topleft = universum._window_start
        self.view_rect.botright = universum._window_end
        self.tmp_drag_offset = np.zeros(2)
        self.drag_start = np.zeros(2)
        self.offset = np.zeros(2)
        self.pause = False
        self.drag_start = np.zeros(2)
        self.zoom = 0.1

    def init_game(self, game):
        self.game = game

    def draw(self, surface: pygame.Surface):
        start = self.view_rect.topleft + self.tmp_drag_offset + self.offset
        end = self.view_rect.botright + self.tmp_drag_offset + self.offset
        self.universum.set_window(start, end)

        if not self.pause:
            for _ in range(5):
                self.universum.simulate(self.game.dt)
        image_buffer = self.universum.image_buffer
        image = pygame.image.frombuffer(image_buffer, image_buffer.shape[:2], "RGB")
        image = pygame.transform.scale(image, self.game.screen_size)
        surface.blit(image, (0, 0))


    def process_event(self, event: pygame.event.Event):
        mouse = np.array(pygame.mouse.get_pos())

        if event.type == pygame.MOUSEBUTTONDOWN:
            # zoom
            if event.button == 4:
                self.view_rect.zoom(-self.zoom)
            elif event.button == 5:
                self.view_rect.zoom(self.zoom)
            
            # drag
            if event.button == 1:
                self.drag_start = mouse
        if event.type == pygame.MOUSEBUTTONUP:
            if event.button == 1:
                self.offset += self.view_rect.diagonal() * 1/1000 * (self.drag_start - mouse)
                self.tmp_drag_offset = np.zeros(2)
                self.drag_start = np.zeros(2)
        elif event.type == pygame.MOUSEMOTION:
            if np.all(self.drag_start != 0):
                self.tmp_drag_offset = self.view_rect.diagonal() * 1/1000 * (self.drag_start - mouse)
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SPACE:
                self.pause = not self.pause
            
            # screenshot
            if event.key == pygame.K_F2:
                self.universum.screenshot("screenshot.png")


def random_universum_factory(width, height, n, dim, type=np.float32):
    candidates_x = np.linspace(0, n, n)
    candidates_y = np.linspace(0, n, n)
    xx = np.random.choice(candidates_x, n)
    yy = np.random.choice(candidates_y, n)
    positions = []
    for x, y in zip(xx, yy):
        positions.append([x, y])
    positions = np.array(positions, type)
    velocities = (np.random.random((n, dim)) - 0.5) * 2 * 2
    masses = np.random.random(n) * 1000
    image_buffer = np.zeros((width, height, 3), type)
    return Universum.from_np_arr(positions, velocities, masses, image_buffer)


if __name__ == "__main__":
    game = Game(1200, 1200)
    game.add_gameobject(UniversObject(random_universum_factory(2000, 2000, 3, 2)))
    game.add_gameobject(UniversObject(random_universum_factory(500, 500, 1000, 2)))
    game.run()